from pypif.obj.common.property import Property

from .base import DFTParser, Value_if_true
import os
from pypif.obj.common.value import Value
from ase import Atoms

class PwscfParser(DFTParser):
    '''
    Parser for PWSCF calculations
    '''
    
    def get_name(self): return "PWSCF"

    def _get_line(self, search_string, search_file, basedir=None, return_string=True, case_sens=True):
        '''Return the first line containing a set of strings in a file.

        If return_string is False, we just return whether such a line
        was found. If case_sens is False, the search is case
        insensitive.

        '''
        if basedir == None: basedir = self._directory # default
        if os.path.isfile(os.path.join(basedir, search_file)):
            # if single search string
            if type(search_string) == type(''): search_string = [search_string]
            # if case insensitive, convert everything to lowercase
            if not case_sens: search_string = [i.lower() for i in search_string]
            with open(os.path.join(basedir, search_file)) as fp:
                # search for the strings line by line
                for line in fp:
                    query_line = line if case_sens else line.lower()
                    if all([i in query_line for i in search_string]):
                        return line if return_string else True
                if return_string:
                    raise Exception('%s not found in %s'%(' & '.join(search_string),os.path.join(basedir, search_file)))
                else: return False
        else: raise Exception('%s file does not exist'%os.path.join(basedir, search_file))
    
    def test_if_from(self, directory):
        '''Look for PWSCF input and output files'''
        self.inputf = self.outputf = ''
        files = [f for f in os.listdir(directory) if os.path.isfile(os.path.join(directory, f))]
        for f in files:
            if self._get_line('Program PWSCF', f, basedir=directory, return_string=False):
                self.outputf = f
            elif self._get_line('&control', f, basedir=directory, return_string=False, case_sens=False):
                self.inputf = f
            if self.inputf and self.outputf: return True
        return False

    def get_version_number(self):
        '''Determine the version number from the output'''
        line = self._get_line('Program PWSCF', self.outputf)
        version = " ".join(line.split('start')[0].split()[2:]).lstrip('v.')
        # return Value(scalars=version)
        return version

    def get_xc_functional(self):
        '''Determine the xc functional from the output'''
        # the xc functional is described by 1 or 4 strings
        # SLA  PZ   NOGX NOGC ( 1 1 0 0 0)
        # SLA  PW   PBX  PBC (1434)
        # PBE ( 1  4  3  4 0 0)
        # PBE0 (6484)
        # HSE (14*4)
        xcstring = self._get_line('Exchange-correlation', self.outputf).split()[2:6]
        for word in range(4):
            if xcstring[word][0] == '(':
                xcstring = xcstring[:word]
                break
        return Value(scalars=" ".join(xcstring))

    def get_cutoff_energy(self):
        '''Determine the cutoff energy from the output'''
        cutoff = self._get_line('kinetic-energy cutoff', self.outputf).split()[3:]
        return Value(scalars=float(cutoff[0]), units=cutoff[1])

    def get_total_energy(self):
        '''Determine the total energy from the output'''
        with open(os.path.join(self._directory, self.outputf)) as fp:
            # reading file backwards in case relaxation run
            for line in reversed(fp.readlines()):
                if "!" in line and "total energy" in line:
                    energy = line.split()[4:]
                    return Property(scalars=float(energy[0]), units=energy[1])
            raise Exception('%s not found in %s'%('! & total energy',os.path.join(self._directory, self.outputf)))

    @Value_if_true
    def is_relaxed(self):
        '''Determine if relaxation run from the output'''
        return self._get_line('Geometry Optimization', self.outputf, return_string=False)

    def _is_converged(self):
        '''Determine if calculation converged; for a relaxation (static) run
        we look for ionic (electronic) convergence in the output'''
        if self.is_relaxed():
            # relaxation run case
            return self._get_line(['End of', 'Geometry Optimization'], self.outputf, return_string=False)
        else:
            # static run case
            return self._get_line('convergence has been achieved', self.outputf, return_string=False)

    def get_KPPRA(self):
        '''Determine the no. of k-points in the BZ (from the input) times the
        no. of atoms (from the output)'''
        # Find the no. of k-points
        fp = open(os.path.join(self._directory, self.inputf)).readlines()
        for l,ll in enumerate(fp):
            if "K_POINTS" in ll:
                # determine the type of input
                if len(ll.split()) > 1:
                    if "gamma" in ll.split()[1].lower():
                        ktype = 'gamma'
                    elif "automatic" in ll.split()[1].lower():
                        ktype = 'automatic'
                    else:
                        ktype = ''
                else: ktype = ''
                if ktype == 'gamma':
                    # gamma point:
                    # K_POINTS {gamma}
                    nk = 1
                elif ktype == 'automatic':
                    # automatic:
                    # K_POINTS automatic
                    #  12 12 1 0 0 0
                    line = [int(i) for i in fp[l+1].split()[0:3]]
                    nk = line[0]*line[1]*line[2]
                else:
                    # manual:
                    # K_POINTS
                    #  3
                    #  0.125  0.125  0.0  1.0
                    #  0.125  0.375  0.0  2.0
                    #  0.375  0.375  0.0  1.0
                    nk = 0
                    for k in range(int(fp[l+1].split()[0])):
                        nk += int(float(fp[l+2+k].split()[3]))
                # Find the no. of atoms
                natoms = int(self._get_line('number of atoms/cell', self.outputf).split()[4])
                return Value(scalars=nk*natoms)
        fp.close()
        raise Exception('%s not found in %s'%('KPOINTS',os.path.join(self._directory, self.inputf)))

    @Value_if_true
    def uses_SOC(self):
        '''Looks for line with "with spin-orbit" in the output'''
        return self._get_line('with spin-orbit', self.outputf, return_string=False)

    def get_pp_name(self):
        '''Determine the pseudopotential names from the output'''
        ppnames = []
        # Find the number of atom types
        natomtypes = int(self._get_line('number of atomic types', self.outputf).split()[5])
        # Find the pseudopotential names
        with open(os.path.join(self._directory, self.outputf)) as fp:
            for line in fp:
                if "PseudoPot. #" in line:
                    ppnames.append(next(fp).split('/')[-1].rstrip())
                    if len(ppnames) == natomtypes:
                        return Value(scalars=ppnames)
            raise Exception('Could not find %i pseudopotential names'%natomtypes)

    def get_U_settings(self):
        '''Determine the DFT+U type and parameters from the output'''
        with open(os.path.join(self._directory, self.outputf)) as fp:
            for line in fp:
                if "LDA+U calculation" in line:
                    U_param = {}
                    U_param['Type'] = line.split()[0]
                    U_param['Values'] = {}
                    # look through next several lines
                    for nl in range(15):
                        line2 = next(fp).split()
                        if len(line2) > 1 and line2[0] == "atomic":
                            pass # column titles
                        elif len(line2) == 6:
                            U_param['Values'][line2[0]] = {}
                            U_param['Values'][line2[0]]['L'] = float(line2[1])
                            U_param['Values'][line2[0]]['U'] = float(line2[2])
                            U_param['Values'][line2[0]]['J'] = float(line2[4])
                        else: break # end of data block
                    return Value(**U_param)
            return None

    def get_vdW_settings(self):
        '''Determine the vdW type if using vdW xc functional or correction
        scheme from the input otherwise'''
        xc = self.get_xc_functional().scalars[0].value
        if 'vdw' in xc.lower(): # vdW xc functional
            return Value(scalars=xc)
        else:
            # look for vdw_corr in input
            vdW_dict = {'xdm':'Becke-Johnson XDM', 'ts':
                        'Tkatchenko-Scheffler', 'ts-vdw':
                        'Tkatchenko-Scheffler',
                        'tkatchenko-scheffler':
                        'Tkatchenko-Scheffler', 'grimme-d2': 'Grimme D2', 'dft-d': 'Grimme D2'}
            if self._get_line('vdw_corr', self.inputf, return_string=False, case_sens=False):
                line = self._get_line('vdw_corr', self.inputf, return_string=True, case_sens=False)
                vdwkey = str(line.split('=')[-1].replace("'", "").replace(',', '').lower().rstrip())
                return Value(scalars=vdW_dict[vdwkey])
            return None

    def get_pressure(self):
        '''Determine the pressure from the output'''
        if self._get_line('total   stress', self.outputf, return_string=False) == False:
            return None
        else:
            line = self._get_line('total   stress', self.outputf)
            # total   stress  (Ry/bohr**3)                   (kbar)     P=   -2.34
            # P= runs into the value if very large pressure
            pvalue = float(line.split()[-1].replace('P=', ''))
            return Property(scalars=pvalue, units='kbar')

    def get_stresses(self):
        '''Determine the stress tensor from the output'''
        stress_data = [] # stress tensor at each iteration
        with open(os.path.join(self._directory, self.outputf)) as fp:
            for line in fp:
                if "total" in line and "stress" in line:
                    stress = []
                    for i in range(3):
                        stress.append([float(j) for j in next(fp).split()[3:6]])
                    stress_data.append(stress)
            if len(stress_data) > 0:
                # return the final stress tensor
                return Property(matrices=stress_data[-1], units='kbar')
            else: return None

    def get_output_structure(self):
        '''Determine the structure from the output'''
        bohr_to_angstrom = 0.529177249

        # determine the number of atoms
        natoms = int(float(self._get_line('number of atoms/cell', self.outputf).split('=')[-1]))

        # determine the initial lattice parameter
        alat = float(self._get_line('lattice parameter (alat)', self.outputf).split('=')[-1].split()[0])

        # find the initial unit cell
        unit_cell = []
        with open(os.path.join(self._directory, self.outputf), 'r') as fp:
            for line in fp:
                if "crystal axes:" in line:
                    for i in range(3):
                        unit_cell.append([float(j)*alat*bohr_to_angstrom for j in next(fp).split('(')[-1].split(')')[0].split()])
                    break
            if len(unit_cell) == 0: raise Exception('Cannot find the initial unit cell')

        # find the initial atomic coordinates
        coords = [] ; atom_symbols = []
        with open(os.path.join(self._directory, self.outputf), 'r') as fp:
            for line in fp:
                if "site n." in line and "atom" in line and "positions" in line and "alat units" in line:
                    for i in range(natoms):
                        coordline = next(fp)
                        atom_symbols.append(''.join([i for i in coordline.split()[1] if not i.isdigit()]))
                        coord_conv_factor = alat*bohr_to_angstrom
                        coords.append([float(j)*coord_conv_factor for j in coordline.rstrip().split('=')[-1].split('(')[-1].split(')')[0].split()])
                    break
            if len(coords) == 0: raise Exception('Cannot find the initial atomic coordinates')

        if type(self.is_relaxed()) == type(None):
            # static run: create, populate, and return the initial structure
            structure = Atoms(symbols=atom_symbols, cell=unit_cell, pbc=True)
            structure.set_positions(coords)
            return structure
        else:
            # relaxation run: update with the final structure
            with open(os.path.join(self._directory, self.outputf)) as fp:
                for line in fp:
                    if "Begin final coordinates" in line:
                        if 'new unit-cell volume' in next(fp):
                            # unit cell allowed to change
                            next(fp) # blank line
                            # get the final unit cell
                            unit_cell = []
                            cellheader = next(fp)
                            if 'bohr' in cellheader.lower():
                                cell_conv_factor = bohr_to_angstrom
                            elif 'angstrom' in cellheader.lower():
                                cell_conv_factor = 1.0
                            else:
                                alat = float(cellheader.split('alat=')[-1].replace(')', ''))
                                cell_conv_factor = alat*bohr_to_angstrom
                            for i in range(3):
                                unit_cell.append([float(j)*cell_conv_factor for j in next(fp).split()])
                            next(fp) # blank line

                        # get the final atomic coordinates
                        coordtype = next(fp).split()[-1].replace('(', '').replace(')', '')
                        if coordtype == 'bohr':
                            coord_conv_factor = bohr_to_angstrom
                        elif coordtype == 'angstrom' or coordtype == 'crystal':
                            coord_conv_factor = 1.0
                        else:
                            coord_conv_factor = alat*bohr_to_angstrom
                        coords = [] # reinitialize the coords
                        for i in range(natoms):
                            coordline = next(fp).split()
                            coords.append([float(j)*coord_conv_factor for j in coordline[1:4]])

                        # create, populate, and return the final structure
                        structure = Atoms(symbols=atom_symbols, cell=unit_cell, pbc=True)
                        if coordtype == 'crystal':
                            structure.set_scaled_positions(coords) # direct coord
                        else:
                            structure.set_positions(coords) # cartesian coord
                        return structure
                raise Exception('Cannot find the final coordinates')

    def get_dos(self):
        '''Find the total DOS shifted by the Fermi energy'''
        # find the dos file
        fildos = ''
        files = [f for f in os.listdir(self._directory) if os.path.isfile(os.path.join(self._directory, f))]
        for f in files:
            fp = open(os.path.join(self._directory, f), 'r')
            first_line = next(fp)
            if "E (eV)" in first_line and "Int dos(E)" in first_line:
                fildos = f
                ndoscol = len(next(fp).split())-2 # number of spin channels
                fp.close()
                break
            fp.close()
        if not fildos: return None # cannot find DOS

        # get the Fermi energy
        line = self._get_line('the Fermi energy is', self.outputf)
        efermi = float(line.split('is')[-1].split()[0])

        # grab the DOS
        energy = [] ; dos = []
        fp = open(os.path.join(self._directory, fildos), 'r')
        next(fp) # comment line
        for line in fp:
            ls = line.split()
            energy.append(float(ls[0])-efermi)
            dos.append(sum([float(i) for i in ls[1:1+ndoscol]]))
        return Property(scalars=dos, units='number of states per unit cell', conditions=Value(name='energy', scalars=energy, units='eV'))

    def get_forces(self):
        return None

    def get_outcar(self):
        return None

    def get_incar(self):
        return None

    def get_poscar(self):
        return None

    def get_band_gap(self):
        '''Compute the band gap from the DOS'''
        dosdata = self.get_dos()
        if type(dosdata) == type(None):
            return None # cannot find DOS
        else:
            energy = dosdata.conditions.scalars
            dos = dosdata.scalars
            step_size = energy[1].value - energy[0].value
            not_found = True ; l = 0 ; bot = 10**3 ; top = -10**3
            while not_found and l < len(dos):
                # iterate through the data
                e = float(energy[l].value)
                dens = float(dos[l].value)
                # note: dos already shifted by efermi
                if e < 0 and dens > 1e-3:
                    bot = e
                elif e > 0 and dens > 1e-3:
                    top = e
                    not_found = False
                l += 1
            if top < bot:
                raise Exception('Algorithm failed to find the band gap')
            elif top - bot < step_size*2:
                return Property(scalars=0, units='eV')
            else:
                bandgap = float(top-bot)
                return Property(scalars=round(bandgap,3), units='eV')



import sys
import json
import logging
import requests
from pypif import pif
from flask import Flask, request
from flask_cors import CORS
from dfttopif import *


# Configure flask
app = Flask(__name__)
CORS(app)

# Configure logging
logging.basicConfig(stream=sys.stdout, level=logging.INFO)


@app.route('/convert/from/tarfile', methods=['POST'])
def convert_from_tarfile():
    # Create a temporary directory to save the files and cleanup when
    # finished with it
    temp_dir_name = '/tmp/' + str(uuid.uuid4())
    os.makedirs(temp_dir_name)
    try:
        data = json.loads(request.get_data(as_text=True))
        response = requests.get(data['url'], stream=True)
        filename = temp_dir_name + '/file_to_process'
        with open(filename, 'wb') as output:
            shutil.copyfileobj(response.raw, output)
        return pif.dumps({'system': tarfile_to_pif(filename, '/tmp/')})
    finally:
        shutil.rmtree(temp_dir_name)



import os
import uuid
import tarfile
import shutil
from dfttopif.parsers import VaspParser
from dfttopif.parsers import PwscfParser
from pypif.obj import *


def tarfile_to_pif(filename, temp_root_dir='', verbose=0):
    """
    Process a tar file that contains DFT data.

    Input:
        filename - String, Path to the file to process.
        temp_root_dir - String, Directory in which to save temporary files. Defaults to working directory.
        verbose - int, How much status messages to print

    Output:
        pif - ChemicalSystem, Results and settings of
            the DFT calculation in pif format
    """
    temp_dir = temp_root_dir + str(uuid.uuid4())
    os.makedirs(temp_dir)
    try:
        tar = tarfile.open(filename, 'r')
        tar.extractall(path=temp_dir)
        tar.close()
        for i in os.listdir(temp_dir):
            cur_dir = temp_dir + '/' + i
            if os.path.isdir(cur_dir):
                return directory_to_pif(cur_dir, verbose)
        return directory_to_pif(temp_dir, verbose)
    finally:
        shutil.rmtree(temp_dir)


def archive_to_pif(filename, verbose=0):
    """
    Given a archive file that contains output from a DFT calculation, parse the data and return a PIF object.

    Input:
        filename - String, Path to the file to process.
        verbose - int, How much status messages to print

    Output:
        pif - ChemicalSystem, Results and settings of
            the DFT calculation in pif format
    """
    if tarfile.is_tarfile(filename):
        return tarfile_to_pif(filename, verbose)
    raise Exception('Cannot process file type')


def directory_to_pif(directory, verbose=0, quality_report=False):
    '''Given a directory that contains output from
    a DFT calculation, parse the data and return
    a pif object

    Input:
        directory - String, path to directory containing
            DFT results
        verbose - int, How much status messages to print

    Output:
        pif - ChemicalSystem, Results and settings of
            the DFT calculation in pif format
    '''

    # Look for the first parser compatible with the directory
    foundParser = False
    for possible_parser in [VaspParser, PwscfParser]:
        try:
            parser = possible_parser(directory)
            if parser.test_if_from(directory):
                foundParser = True
                break
        except: pass
    if not foundParser:
        raise Exception('Directory is not in correct format for an existing parser')
    if verbose > 0:
        print("Found a %s directory", parser.get_name())
        
    # Get information about the chemical system
    chem = ChemicalSystem()
    chem.chemical_formula = parser.get_composition()
        
    # Get software information, to list as method
    software = Software(name=parser.get_name(),
        version=parser.get_version_number())
        
    # Define the DFT method object
    method = Method(name='Density Functional Theory',
        software=software)
        
    # Get the settings (aka. "conditions") of the DFT calculations
    conditions = []
    for name, func in parser.get_setting_functions().items():
        # Get the condition
        cond = getattr(parser, func)()

        # If the condition is None or False, skip it
        if cond is None:
            continue

        # Set the name
        cond.name = name

        # Set the types
        conditions.append(cond)
    
    # Get the properties of the system
    chem.properties = []
    for name, func in parser.get_result_functions().items():
        # Get the property
        prop = getattr(parser, func)()
        
        # If the property is None, skip it
        if prop is None:
            continue

        # Add name and other data
        prop.name = name
        prop.method = method
        prop.data_type='COMPUTATIONAL'
        if verbose > 0 and isinstance(prop, Value):
            print(name)
        if prop.conditions is None:
            prop.conditions = conditions
        else:
            if not isinstance(prop.conditions, list):
                prop.conditions = [prop.conditions]
            prop.conditions.extend(conditions)

        # Add it to the output
        chem.properties.append(prop)

    if quality_report:
        import tarfile
        tar = tarfile.open("tmp.tar", "w")
        tar.add(os.path.join(directory, "OUTCAR"))
        tar.add(os.path.join(directory, "INCAR"))
        tar.close()

        import requests
        import json
        r = requests.post('https://calval.citrination.com/validate/tarfile', data=open('tmp.tar', 'rb').read())

        if r.status_code == requests.codes.ok:
            report = r.json()[0]
            report_file = os.path.join(directory, "report.txt")
            with open(report_file, "w") as f:
                f.write(report)
            if report_file[0:2] == "./":
                report_file = report_file[2:]
            chem.properties.append(
                    Property(
                        name="quality_report",
                        files=[FileReference(relative_path=report_file)]
                    )
                )
        else:
            print("Something failed: {}".format(r.status_code))
            print(r.status_code)


    return chem



from .base import DFTParser
import os
import glob
from ase.calculators.abinit import Abinit
from pypif.obj.common import Value, Property

class AbinitParser(DFTParser):
    '''
    Parser for ABINIT calculations
    '''
    _label = None
    
    def get_name(self): return "ABINIT"
    
    def test_if_from(self, directory):
        # Check whether any file has as name ABINIT in the file in the first two lines
        files = [f for f in os.listdir(directory) if os.path.isfile(os.path.join(directory, f))]
        for f in files:
            fp = open(os.path.join(directory, f), 'r')
            for line in [fp.readline(), fp.readline()]:
                if "ABINIT" in line:
                    fp.close()
                    return True
            fp.close()
        return False
        
    def _get_label(self):
        '''Find the label for the output files 
         for this calculation
        '''
        if self._label is None:
            files = [f for f in os.listdir(self._directory) if os.path.isfile(os.path.join(self._directory, f))] 
            foundfiles = False
            print(files)
            for f in files:
                if ".files" in f: 
                    foundfiles = True
                    self._label = f.split(".")[0]
                    fp = open(os.path.join(self._directory, self._label + '.files'), 'r')
                    line = fp.readline().split()[0]
                    if line != self._label + ".in":
                       fp.close()
                       raise Exception('first line must be label.in') 
                    line = fp.readline().split()[0]
                    if line != self._label + ".txt":
                       fp.close()
                       raise Exception('second line must be label.txt') 
                    line = fp.readline().split()[0]
                    if line != self._label + "i":
                       fp.close()
                       raise Exception('third line must be labeli') 
                    line = fp.readline().split()[0]
                    if line != self._label + "o":
                       fp.close()
                       raise Exception('fourth line must be labelo') 
                    fp.close()
            if foundfiles:
                return self._label
            else:
                raise Exception('label.files not found')
                                
#ASE format
#        (self.prefix + '.in') # input
#        (self.prefix + '.txt')# output
#        (self.prefix + 'i')   # input
#        (self.prefix + 'o')   # output
                
        else:
            return self._label

    def get_cutoff_energy(self):
        # Open up the label.txt file
        fp = open(os.path.join(self._directory, self._label + '.out'), 'r')
        foundecho = False 
        # Look for ecut after the occurence of "echo values of preprocessed input variables"
        for line in fp:
            if "echo values of preprocessed input variables" in line:
                foundecho = True
            if "ecut" in line and foundecho:
                words = line.split()
                return Value(scalars=float(words[1]), units=words[2])
        # Error handling: ecut not found
        raise Exception('ecut not found')



from pypif.obj.common.property import Property

from .base import DFTParser, Value_if_true
import os
from ase.calculators.vasp import Vasp
from ase.io.vasp import read_vasp, read_vasp_out
from pypif.obj.common.value import Value
from pypif.obj.common.file_reference import FileReference

class VaspParser(DFTParser):
    '''
    Parser for VASP calculations
    '''

    def get_name(self): return "VASP"
    
    def test_if_from(self, directory):
        # Check whether it has an INCAR file
        return os.path.isfile(os.path.join(directory, 'INCAR'))
        
    def get_output_structure(self):
        self.atoms = read_vasp_out(os.path.join(self._directory, 'OUTCAR'))
        return self.atoms

    def get_outcar(self):
        raw_path = os.path.join(self._directory, 'OUTCAR')
        if raw_path[0:2] == "./":
            raw_path = raw_path[2:]
        return Property(files=[FileReference(
            relative_path=raw_path
        )])

    def get_incar(self):
        raw_path = os.path.join(self._directory, 'INCAR')
        if raw_path[0:2] == "./":
            raw_path = raw_path[2:]
        return Value(files=[FileReference(
            relative_path=raw_path
        )])

    def get_poscar(self):
        raw_path = os.path.join(self._directory, 'POSCAR')
        if raw_path[0:2] == "./":
            raw_path = raw_path[2:]
        return Value(files=[FileReference(
            relative_path=raw_path
        )])

    def get_cutoff_energy(self):
        # Open up the OUTCAR
        with open(os.path.join(self._directory, 'OUTCAR'), 'r') as fp:
         # Look for ENCUT
            for line in fp:
                if "ENCUT" in line:
                    words = line.split()
                    return Value(scalars=float(words[2]), units=words[3])
                
        # Error handling: ENCUT not found
        raise Exception('ENCUT not found')

    @Value_if_true
    def uses_SOC(self):
        # Open up the OUTCAR
        with open(os.path.join(self._directory, 'OUTCAR')) as fp:
        
            #look for LSORBIT
            for line in fp:
                if "LSORBIT" in line:
                    words = line.split()
                    return words[2] == 'T'
        
        # Error handling: LSORBIT not found
        raise Exception('LSORBIT not found')
        
    @Value_if_true
    def is_relaxed(self):
        # Open up the OUTCAR
        with open(os.path.join(self._directory, 'OUTCAR')) as fp:
        
            #  Look for NSW
            for line in fp:
                if "NSW" in line:
                    words = line.split()
                    return int(words[2]) != 0
                
        # Error handling: NSW not found
        raise Exception('NSW not found')
        
    def get_xc_functional(self):
        # Open up the OUTCAR
        with open(os.path.join(self._directory, 'OUTCAR')) as fp:
        
            # Look for TITEL
            for line in fp:
                if "TITEL" in line:
                    words = line.split()
                    return Value(scalars=words[2])
                    break
            
    def get_pp_name(self):
        # Open up the OUTCAR
        with open(os.path.join(self._directory, 'OUTCAR')) as fp:
        
            #initialize empty list to store pseudopotentials
            pp = []
            # Look for TITEL
            for line in fp:
                if "TITEL" in line:
                    words = line.split()
                    pp.append(words[3])
            return Value(scalars=pp)
                
        # Error handling: TITEL not found
        raise Exception('TITEL not found')
        
    def get_KPPRA(self):
        # Open up the OUTCAR
        with open(os.path.join(self._directory, 'OUTCAR')) as fp:
            #store the number of atoms and number of irreducible K-points
            for line in fp:
                if "NIONS" in line:
                    words = line.split()
                    NI = int(words[11])
                elif "NKPTS" in line:
                    words = line.split()
                    NIRK = float(words[3])
            #check if the number of k-points was reduced by VASP if so, sum all the k-points weight
            if "irreducible" in open(os.path.join(self._directory, 'OUTCAR')).read():
                fp.seek(0)
                for line in fp:
                    #sum all the k-points weight
                    if "Coordinates               Weight" in line:
                        NK=0; counter = 0
                        for line in fp:
                            if counter == NIRK:
                                break
                            NK += float(line.split()[3])
                            counter += 1
                return Value(scalars=(NI*NK))
            #if k-points were not reduced KPPRA equals the number of atoms * number of irreducible k-points
            else:
                return Value(scalars=(NI*NIRK))

                
        # Error handling: NKPTS or NIONS not found
        raise Exception('NIONS, irredicuble or Coordinates not found')

    def _is_converged(self):
        return self._call_ase(Vasp().read_convergence)

    def get_total_energy(self):
        return Property(scalars=self._call_ase(Vasp().read_energy)[0], units='eV')

    def get_version_number(self):
        # Open up the OUTCAR
        with open(os.path.join(self._directory, 'OUTCAR')) as fp:
        
            #look for vasp
            for line in fp:
                if "vasp" in line:
                    words = line.split()
                    return (words[0].strip('vasp.'))
                    break
        
        # Error handling: vasp not found
        raise Exception('vasp not found')
        
    def get_U_settings(self):
        #Open up the OUTCAR
        with open(os.path.join(self._directory, 'OUTCAR')) as fp:
            #Check if U is used
            if "LDAU" in open(os.path.join(self._directory, 'OUTCAR')).read():
                U_param = {}
                atoms = []
                #get the list of pseupotential used
                for line in fp:
                    if "TITEL" in line:
                        atoms.append(line.split()[3])
                    #Get the U type used
                    if "LDAUTYPE" in line:
                            U_param['Type'] = int(line.split()[-1])
                atoms.reverse()
                fp.seek(0)
                #Get the L value
                U_param['Values'] = {}
                for line in fp:
                    for atom, i in zip(atoms, range(len(atoms))):
                        if "LDAUL" in line:
                                U_param['Values'][atom] = {'L': int(line.split()[-1-i])}
                fp.seek(0)
                #Get the U value            
                for line in fp:
                    for atom, i in zip(atoms, range(len(atoms))):
                        if "LDAUU" in line:
                                U_param['Values'][atom]['U'] = float(line.split()[-1-i])
                fp.seek(0)
                #Get the J value
                for line in fp:
                    for atom, i in zip(atoms, range(len(atoms))):
                        if "LDAUJ" in line:
                                U_param['Values'][atom]['J'] = float(line.split()[-1-i])
                return Value(**U_param)
            #if U is not used, return None
            else:
                return None
            
    def get_vdW_settings(self):
        #define the name of the vdW methods in function of their keyword
        vdW_dict = {'BO':'optPBE-vdW', 'MK':'optB88-vdW', 'ML':'optB86b-vdW','RE':'vdW-DF','OR':'Klimes-Bowler-Michaelides'}
        #Open up the OUTCAR
        with open(os.path.join(self._directory, 'OUTCAR')) as fp:
            #Check if vdW is used
            if "LUSE_VDW" in open(os.path.join(self._directory, 'OUTCAR')).read():
                #if vdW is used, get its keyword             
                for line in fp:
                    if "GGA     =" in line:
                        words = line.split()
                        return Value(scalars=vdW_dict[words[2]])
            #if vdW is not used, return None
            else:
                return None
            
    def get_pressure(self):
        #define pressure dictionnary because since when is kB = kbar? Come on VASP people
        pressure_dict = {'kB':'kbar'}
        #Check if ISIF = 0 is used
        if "ISIF   =      0" in open(os.path.join(self._directory, 'OUTCAR')).read():
            #if ISIF = 0 is used, print this crap
            return None
        #if ISIF is not 0 then extract pressure and units
        else:
            #scan file in reverse to have the final pressure
            for line in reversed(open(os.path.join(self._directory, 'OUTCAR')).readlines()):
                if "external pressure" in line:
                    words = line.split()
                    return Property(scalars=float(words[3]), units=pressure_dict[words[4]])
                    break
    
    def get_stresses(self):
        #Check if ISIF = 0 is used
        if "ISIF   =      0" in open(os.path.join(self._directory, 'OUTCAR')).read():
            return None
        #Check if ISIF = 1 is used
        elif "ISIF   =      1" in open(os.path.join(self._directory, 'OUTCAR')).read():
            return None
        else:
            #scan file in reverse to have the final pressure
            for line in open(os.path.join(self._directory, 'OUTCAR')).readlines():
                if "in kB" in line:
                    words = line.split()
                    XX = float(words[2]); YY = float(words[3]); ZZ = float(words[4]); XY= float(words[5]); YZ = float(words[6]); ZX = float(words[7])
            return Property(matrices=[[XX,XY,ZX],[XY,YY,YZ],[ZX,YZ,ZZ]], units='kbar')
            
         # Error handling: "in kB" not found
        raise Exception('in kB not found')

    def get_forces(self):
        self.atoms = read_vasp_out(os.path.join(self._directory, 'OUTCAR'))
        return Property(
            vectors=self.atoms.get_calculator().results['forces'].tolist(),
            conditions=Value(name="positions", vectors=self.atoms.positions.tolist())
        )



    def get_band_gap(self):
        file_path = os.path.join(self._directory, 'DOSCAR')
        if not os.path.isfile(file_path):
            return None
        #open DOSCAR
        with open(os.path.join(self._directory, 'DOSCAR')) as fp:
            for i in range(6):
                l = fp.readline()
            efermi = float(l.split()[3])
            step1 = fp.readline().split()[0]
            step2 = fp.readline().split()[0]
            step_size = float(step2)-float(step1)
            not_found = True
            while not_found:
                l = fp.readline().split()
                e = float(l.pop(0))
                dens = 0.0
                for i in range(int(len(l)/2)):
                    dens += float(l[i])
                if e < efermi and dens > 1e-3:
                    bot = e
                elif e > efermi and dens > 1e-3:
                    top = e
                    not_found = False
            if top - bot < step_size*2:
                return Property(scalars=0, units='eV')
            else:
                bandgap = float(top - bot)
                return Property(scalars=round(bandgap,3), units='eV')
                
    def get_dos(self):
        file_path = os.path.join(self._directory, 'DOSCAR')
        if not os.path.isfile(file_path):
            return None
        #open DOSCAR
        with open(os.path.join(self._directory, 'DOSCAR')) as fp:
            for i in range(6):
                l = fp.readline()
            n_step = int(l.split()[2])
            energy = []; dos = []            
            for i in range(n_step):
                l = fp.readline().split()
                e = float(l.pop(0))
                energy.append(e)
                dens = 0
                for j in range(int(len(l)/2)):
                    dens += float(l[j])
                dos.append(dens)

            # Convert to property
            return Property(scalars=dos, units='number of states per unit cell',
                            conditions=Value(name='energy', scalars=energy, units='eV'))



import os
from collections import Counter
from pypif.obj.common import Value, Property


def Value_if_true(func):
    '''Returns:
        Value if x is True, else None'''
    return lambda x: Value() if func(x) == True else None


class DFTParser(object):
    '''Base class for all tools to parse a directory of output files from a DFT Calculation
    
    To use this class, provide the path to a directory to containing the output files from a DFT calculation. Once
     instantiated, call the methods provided by this class in order to retrieve the settings and results of the
     calculation.
    
    To get a list of names of the settings available for this a particular instance, call get_setting_functions(). These
     methods return a pypif Value object.
    
    To get a list of the names of results available via a particular instance, call get_result_functions(). These
     methods return a pypif Property object.

    Developer Notes
    ---------------
    
    Settings and properties should...
        return None if property or setting is not applicable, or if the test is a boolean and the value is false
        Raise exception if there is an error that would benefit from user intervention

    To add a new setting or value to the output, add a new entry to the dictionary returned by get_setting_functions()
     or get_result_functions(). The key for each entry is a 'human-friendly' name of the result and the value is the
     name of the function. This design was chosen because there is a single function for defining the human names of the
     results, which are what serve as the tags in the pif file. In this way, the same property will be ensured to
     have the same name in the pif.
    '''
    
    _directory = None
    '''Path to directory containing calculation files'''
    
    _converged = None
    ''' Whether this calculation has converged '''
    
    def __init__(self, directory):
        '''Initialize a parser.
        
        Input:
            directory - String, path to a directory of output files
        '''
        
        # Sanity check: Make sure the format is correct
        if not self.test_if_from(directory):
            raise Exception('Files in directory inconsistent with this format')
            
        self._directory = directory
      
    @classmethod
    def test_if_from(self, directory):
        '''Test whether a directory of output files
        seems like it is from this DFT code.
        
        Input:
            directory - String, path to a directory of output files
        Returns: 
            boolean, whether directory was created by this code
        '''
        
        raise NotImplementedError
        
    def get_setting_functions(self):
        '''Get a dictionary containing the names of methods
        that return settings of the calculation
        
        Returns:
            dict, where the key is the name of the setting,
                and the value is function name of this parser
        '''
        return {
            'XC Functional':'get_xc_functional',
            'Relaxed':'is_relaxed',
            'Cutoff Energy':'get_cutoff_energy',
            'k-Points per Reciprocal Atom':'get_KPPRA',
            'Spin-Orbit Coupling':'uses_SOC',
            'DFT+U':'get_U_settings',
            'vdW Interactions':'get_vdW_settings',
            'Psuedopotentials':'get_pp_name',
            'INCAR':'get_incar',
            'POSCAR':'get_poscar',
        }
        
    def get_result_functions(self):
        '''Get a dictionary describing the names of methods 
        that return results of the calculation
        
        Returns:
            dict, where the key is the name of a property,
                and the value is the name of the function
        '''
        return {
            'Converged':'is_converged',
            'Total Energy':'get_total_energy',
            'Band Gap Energy':'get_band_gap',
            'Pressure':'get_pressure',
            'Density of States':'get_dos',
            'Positions': 'get_positions',
            'Forces': 'get_forces',
            'OUTCAR': 'get_outcar',
        }
        
    def _call_ase(self, func):
        '''Make a call to an ASE function.
        
        Handles changing directories
        
        Returns: Result of ASE function
        '''
        # Change directories
        old_path = os.getcwd()
        os.chdir(self._directory)
        
        # Call function
        try:
            res = func()
        except:
            os.chdir(old_path)
            raise
        
        # Change back
        os.chdir(old_path)
        return res
        
    def get_name(self):
        '''Get the name of this program'''
        raise NotImplementedError
        
    def get_version_number(self):
        '''Get the version number of code that
        created these output files
        
        Returns:
            string, Version number
        '''
        raise NotImplementedError
    
    def get_output_structure(self):
        '''Get the output structure, if available
        
        Returns:
            ase.Atoms - Output structure from this calculation
                or None if output file not found
        '''
        raise NotImplementedError
    
    def get_composition(self):
        '''Get composition of output structure
        
        Returns:
            String - Composition based on output structure
        '''
        strc = self.get_output_structure()
        counts = Counter(strc.get_chemical_symbols())
        return ''.join(k if counts[k]==1 else '%s%d'%(k,counts[k]) \
                for k in sorted(counts))

    def get_positions(self):
        strc = self.get_output_structure()
        return Property(vectors=strc.positions.tolist())

    def get_cutoff_energy(self):
        '''Read the cutoff energy from the output
        
        Returns:
            Value, cutoff energy (scalar) and units
        '''
        
        raise NotImplementedError
    
    def uses_SOC(self):
        '''Parse the output file to tell if spin-orbit coupling was used
        
        Returns:
            Blank Value if true, `None` otherwise
        '''
        
        raise NotImplementedError
        
    def is_relaxed(self):
        '''Parse the output file to tell if the structure was relaxed
        
        Returns:
            Blank Value if true, `None` otherwise
        '''
        
        raise NotImplementedError
    
    def get_xc_functional(self):
        '''Parse the output file to tell which exchange-correlation functional was used
        
        Returns:
            Value - where "scalars" is the name of the functional
        '''
        
        raise NotImplementedError

    def is_converged(self):
        '''Whether the calculation has converged
        
        Returns: Property where "scalar" is a boolean indicating
        '''

        # Check for cached result
        if self._converged is None:
            self._converged = self._is_converged()
        return Property(scalars=self._converged)

    def get_pp_name(self):
        '''Read output to get the pseudopotentials names used for each elements
        
        Returns:
            Value where the key "scalars" is the list of pseudopotentials names
        '''
        
        raise NotImplementedError

    def get_KPPRA(self):
        '''Read output and calculate the number of k-points per reciprocal atom
        
        Returns:
            Value, number of k-points per reciprocal atom
        '''
        
        raise NotImplementedError

    def get_U_settings(self):
        '''Get the DFT+U settings, if used

        Returns: Value, which could contain several keys
            'Type' -> String, type of DFT+U employed
            'Values' -> dict of Element -> (L, U, J)
        Note: Returns None if DFT+U was not used
        '''

        raise NotImplementedError

    def get_vdW_settings(self):
        '''Get the vdW settings, if applicable

        Returns: Value where `scalars` is the name of the vdW method. None if vdW was not used'''

        raise NotImplementedError

    # Operations for retrieving results
    def _is_converged(self):
        '''Read output to see whether it is converged
        
        Hidden operation: self.is_converged() is the public
        interface, which may draw from a converged result
        
        Returns: boolean'''
        
        raise NotImplementedError
        
    def get_total_energy(self):
        '''Get the total energy of the last ionic step
        
        Returns: Property
        '''
        
        raise NotImplementedError
        
    def get_band_gap(self):
        '''Get the band gap energy

        Returns: Property'''

        raise NotImplementedError

    def get_pressure(self):
        '''Get the pressure acting on the system

        Returns: Property, where pressure is a scalar'''

        raise NotImplementedError

    def get_dos(self):
        '''Get the total density of states

        Returns: Property where DOS is a vector, and the energy at which the DOS was evaluated is a condition'''
            
        raise NotImplementedError

    def get_stresses(self):
       '''Get the stress tensor

       Returns: Property where stresses is a 2d matrix'''

       raise NotImplementedError



from setuptools import setup, find_packages

setup(
    name='dfttopif',
    version='0.1.6',
    description='Library for parsing Density Functional Theory calculations',
    url='https://github.com/CitrineInformatics/pif-dft',
    install_requires=[
        'ase',
        'pypif',
    ],
    packages=find_packages(exclude=('tests', 'docs'))
)
